@@ -1,3 +1,24 @@
+/*
+ * cryptsetup - setup cryptographic volumes for dm-crypt
+ *
+ * Copyright (C) 2004, Christophe Saout <christophe@saout.de>
+ * Copyright (C) 2004-2007, Clemens Fruhwirth <clemens@endorphin.org>
+ * Copyright (C) 2009-2011, Red Hat, Inc. All rights reserved.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License
+ * version 2 as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ */
+
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -6,32 +27,33 @@
 #include <inttypes.h>
 #include <errno.h>
 #include <unistd.h>
+#include <sys/stat.h>
 #include <fcntl.h>
 #include <assert.h>
-
+#include <limits.h>
 #include <libcryptsetup.h>
 #include <popt.h>
 
-#include "../config.h"
-
 #include "cryptsetup.h"
 
 static int opt_verbose = 0;
 static int opt_debug = 0;
-static char *opt_cipher = NULL;
-static char *opt_hash = NULL;
+static const char *opt_cipher = NULL;
+static const char *opt_hash = NULL;
 static int opt_verify_passphrase = 0;
-static char *opt_key_file = NULL;
-static char *opt_master_key_file = NULL;
-static char *opt_header_backup_file = NULL;
-static char *opt_uuid = NULL;
-static unsigned int opt_key_size = 0;
-static unsigned int opt_keyfile_size = 0;
-static unsigned int opt_new_keyfile_size = 0;
+static const char *opt_key_file = NULL;
+static const char *opt_master_key_file = NULL;
+static const char *opt_header_backup_file = NULL;
+static const char *opt_uuid = NULL;
+static const char *opt_header_device = NULL;
+static int opt_key_size = 0;
+static long opt_keyfile_size = 0;
+static long opt_new_keyfile_size = 0;
 static int opt_key_slot = CRYPT_ANY_SLOT;
 static uint64_t opt_size = 0;
 static uint64_t opt_offset = 0;
 static uint64_t opt_skip = 0;
+static int opt_skip_valid = 0;
 static int opt_readonly = 0;
 static int opt_iteration_time = 1000;
 static int opt_batch_mode = 0;
@@ -42,6 +64,8 @@ static int opt_align_payload = 0;
 static int opt_random = 0;
 static int opt_urandom = 0;
 static int opt_dump_master_key = 0;
+static int opt_shared = 0;
+static int opt_allow_discards = 0;
 
 static const char **action_argv;
 static int action_argc;
@@ -55,6 +79,7 @@ static int action_luksOpen(int arg);
 static int action_luksAddKey(int arg);
 static int action_luksKillSlot(int arg);
 static int action_luksRemoveKey(int arg);
+static int action_luksChangeKey(int arg);
 static int action_isLuks(int arg);
 static int action_luksUUID(int arg);
 static int action_luksDump(int arg);
@@ -62,6 +87,7 @@ static int action_luksSuspend(int arg);
 static int action_luksResume(int arg);
 static int action_luksBackup(int arg);
 static int action_luksRestore(int arg);
+static int action_loopaesOpen(int arg);
 
 static struct action_type {
 	const char *type;
@@ -80,18 +106,22 @@ static struct action_type {
 	{ "luksOpen",	action_luksOpen,	0, 2, 1, N_("<device> <name> "), N_("open LUKS device as mapping <name>") },
 	{ "luksAddKey",	action_luksAddKey,	0, 1, 1, N_("<device> [<new key file>]"), N_("add key to LUKS device") },
 	{ "luksRemoveKey",action_luksRemoveKey,	0, 1, 1, N_("<device> [<key file>]"), N_("removes supplied key or key file from LUKS device") },
+	{ "luksChangeKey",action_luksChangeKey,	0, 1, 1, N_("<device> [<key file>]"), N_("changes supplied key or key file of LUKS device") },
 	{ "luksKillSlot",  action_luksKillSlot, 0, 2, 1, N_("<device> <key slot>"), N_("wipes key with number <key slot> from LUKS device") },
 	{ "luksUUID",	action_luksUUID,	0, 1, 0, N_("<device>"), N_("print UUID of LUKS device") },
 	{ "isLuks",	action_isLuks,		0, 1, 0, N_("<device>"), N_("tests <device> for LUKS partition header") },
 	{ "luksClose",	action_remove,		0, 1, 1, N_("<name>"), N_("remove LUKS mapping") },
-	{ "luksDump",	action_luksDump,	0, 1, 0, N_("<device>"), N_("dump LUKS partition information") },
+	{ "luksDump",	action_luksDump,	0, 1, 1, N_("<device>"), N_("dump LUKS partition information") },
 	{ "luksSuspend",action_luksSuspend,	0, 1, 1, N_("<device>"), N_("Suspend LUKS device and wipe key (all IOs are frozen).") },
 	{ "luksResume",	action_luksResume,	0, 1, 1, N_("<device>"), N_("Resume suspended LUKS device.") },
 	{ "luksHeaderBackup",action_luksBackup,	0, 1, 1, N_("<device>"), N_("Backup LUKS device header and keyslots") },
 	{ "luksHeaderRestore",action_luksRestore,0,1, 1, N_("<device>"), N_("Restore LUKS device header and keyslots") },
+	{ "loopaesOpen",action_loopaesOpen,	0, 2, 1, N_("<device> <name> "), N_("open loop-AES device as mapping <name>") },
+	{ "loopaesClose",action_remove,		0, 1, 1, N_("<name>"), N_("remove loop-AES mapping") },
 	{ NULL, NULL, 0, 0, 0, NULL, NULL }
 };
 
+__attribute__((format(printf, 5, 6)))
 static void clogger(struct crypt_device *cd, int level, const char *file,
 		   int line, const char *format, ...)
 {
@@ -116,7 +146,7 @@ static void clogger(struct crypt_device *cd, int level, const char *file,
 	free(target);
 }
 
-static int _yesDialog(const char *msg, void *usrptr)
+static int _yesDialog(const char *msg, void *usrptr __attribute__((unused)))
 {
 	char *answer = NULL;
 	size_t size = 0;
@@ -138,7 +168,7 @@ static int _yesDialog(const char *msg, void *usrptr)
 	return r;
 }
 
-static void _log(int level, const char *msg, void *usrptr)
+static void _log(int level, const char *msg, void *usrptr __attribute__((unused)))
 {
 	switch(level) {
 
@@ -152,12 +182,23 @@ static void _log(int level, const char *msg, void *usrptr)
 	case CRYPT_LOG_ERROR:
 		fputs(msg, stderr);
 		break;
+	case CRYPT_LOG_DEBUG:
+		if (opt_debug)
+			printf("# %s\n", msg);
+		break;
 	default:
 		fprintf(stderr, "Internal error on logging class for msg: %s", msg);
 		break;
 	}
 }
 
+static void _quiet_log(int level, const char *msg, void *usrptr)
+{
+	if (!opt_verbose && (level == CRYPT_LOG_ERROR || level == CRYPT_LOG_NORMAL))
+		level = CRYPT_LOG_VERBOSE;
+	_log(level, msg, usrptr);
+}
+
 static void show_status(int errcode)
 {
 	char error[256], *error_;
@@ -187,7 +228,7 @@ static void show_status(int errcode)
 		log_err(".\n");
 }
 
-static int action_create(int arg)
+static int action_create(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	char cipher[MAX_CIPHER_LEN], cipher_mode[MAX_CIPHER_LEN];
@@ -195,10 +236,12 @@ static int action_create(int arg)
 		.hash = opt_hash ?: DEFAULT_PLAIN_HASH,
 		.skip = opt_skip,
 		.offset = opt_offset,
+		.size = opt_size,
 	};
 	char *password = NULL;
-	unsigned int passwordLen;
-	unsigned int key_size = (opt_key_size ?: DEFAULT_PLAIN_KEYBITS) / 8;
+	size_t passwordLen;
+	size_t key_size = (opt_key_size ?: DEFAULT_PLAIN_KEYBITS) / 8;
+	uint32_t activate_flags = 0;
 	int r;
 
 	if (params.hash && !strcmp(params.hash, "plain"))
@@ -208,8 +251,12 @@ static int action_create(int arg)
 	if (opt_key_file && strcmp(opt_key_file, "-"))
 		params.hash = NULL;
 
+	if (opt_keyfile_size && opt_key_file)
+		log_std(("Ignoring keyfile size option, keyfile read size "
+			 "is always the same as encryption key size.\n"));
+
 	r = crypt_parse_name_and_mode(opt_cipher ?: DEFAULT_CIPHER(PLAIN),
-				      cipher, cipher_mode);
+				      cipher, NULL, cipher_mode);
 	if (r < 0) {
 		log_err("No known cipher specification pattern detected.\n");
 		goto out;
@@ -229,22 +276,31 @@ static int action_create(int arg)
 	if (r < 0)
 		goto out;
 
+	if (opt_readonly)
+		activate_flags |= CRYPT_ACTIVATE_READONLY;
+
+	if (opt_shared)
+		activate_flags |= CRYPT_ACTIVATE_SHARED;
+
+	if (opt_allow_discards)
+		activate_flags |= CRYPT_ACTIVATE_ALLOW_DISCARDS;
+
 	if (opt_key_file)
+		/* With hashing, read the whole keyfile */
 		r = crypt_activate_by_keyfile(cd, action_argv[0],
-			CRYPT_ANY_SLOT, opt_key_file, key_size,
-			opt_readonly ?  CRYPT_ACTIVATE_READONLY : 0);
+			CRYPT_ANY_SLOT, opt_key_file, params.hash ? 0 : key_size,
+			activate_flags);
 	else {
 		r = crypt_get_key(_("Enter passphrase: "),
-				  &password, &passwordLen, 0, NULL,
-				  opt_timeout,
+				  &password, &passwordLen, opt_keyfile_size,
+				  NULL, opt_timeout,
 				  opt_batch_mode ? 0 : opt_verify_passphrase,
 				  cd);
 		if (r < 0)
 			goto out;
 
 		r = crypt_activate_by_passphrase(cd, action_argv[0],
-			CRYPT_ANY_SLOT, password, passwordLen,
-			 opt_readonly ?  CRYPT_ACTIVATE_READONLY : 0);
+			CRYPT_ANY_SLOT, password, passwordLen, activate_flags);
 	}
 out:
 	crypt_free(cd);
@@ -253,7 +309,46 @@ out:
 	return r;
 }
 
-static int action_remove(int arg)
+static int action_loopaesOpen(int arg __attribute__((unused)))
+{
+	struct crypt_device *cd = NULL;
+	struct crypt_params_loopaes params = {
+		.hash = opt_hash ?: NULL,
+		.offset = opt_offset,
+		.skip = opt_skip_valid ? opt_skip : opt_offset,
+	};
+	unsigned int key_size = (opt_key_size ?: DEFAULT_LOOPAES_KEYBITS) / 8;
+	uint32_t activate_flags = 0;
+	int r;
+
+	if (!opt_key_file) {
+		log_err(_("Option --key-file is required.\n"));
+		return -EINVAL;
+	}
+
+	if (opt_readonly)
+		activate_flags |= CRYPT_ACTIVATE_READONLY;
+
+	if (opt_allow_discards)
+		activate_flags |= CRYPT_ACTIVATE_ALLOW_DISCARDS;
+
+	if ((r = crypt_init(&cd, action_argv[0])))
+		goto out;
+
+	r = crypt_format(cd, CRYPT_LOOPAES, opt_cipher ?: DEFAULT_LOOPAES_CIPHER,
+			 NULL, NULL, NULL, key_size, &params);
+	if (r < 0)
+		goto out;
+
+	r = crypt_activate_by_keyfile(cd, action_argv[1], CRYPT_ANY_SLOT,
+				      opt_key_file, opt_keyfile_size, activate_flags);
+out:
+	crypt_free(cd);
+
+	return r;
+}
+
+static int action_remove(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r;
@@ -266,12 +361,12 @@ static int action_remove(int arg)
 	return r;
 }
 
-static int action_resize(int arg)
+static int action_resize(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r;
 
-	r = crypt_init_by_name(&cd, action_argv[0]);
+	r = crypt_init_by_name_and_header(&cd, action_argv[0], opt_header_device);
 	if (r == 0)
 		r = crypt_resize(cd, action_argv[0], opt_size);
 
@@ -279,30 +374,45 @@ static int action_resize(int arg)
 	return r;
 }
 
-static int action_status(int arg)
+static int action_status(int arg __attribute__((unused)))
 {
 	crypt_status_info ci;
 	struct crypt_active_device cad;
 	struct crypt_device *cd = NULL;
-	int r = 0;
+	struct stat st;
+	char *backing_file;
+	const char *device;
+	int path = 0, r = 0;
+
+	/* perhaps a path, not a dm device name */
+	if (strchr(action_argv[0], '/') && !stat(action_argv[0], &st))
+		path = 1;
 
 	ci = crypt_status(NULL, action_argv[0]);
 	switch (ci) {
 	case CRYPT_INVALID:
-		r = -ENODEV;
+		r = -EINVAL;
 		break;
 	case CRYPT_INACTIVE:
-		log_std("%s/%s is inactive.\n", crypt_get_dir(), action_argv[0]);
+		if (path)
+			log_std("%s is inactive.\n", action_argv[0]);
+		else
+			log_std("%s/%s is inactive.\n", crypt_get_dir(), action_argv[0]);
+		r = -ENODEV;
 		break;
 	case CRYPT_ACTIVE:
 	case CRYPT_BUSY:
-		log_std("%s/%s is active%s.\n", crypt_get_dir(), action_argv[0],
-			ci == CRYPT_BUSY ? " and is in use" : "");
+		if (path)
+			log_std("%s is active%s.\n", action_argv[0],
+				ci == CRYPT_BUSY ? " and is in use" : "");
+		else
+			log_std("%s/%s is active%s.\n", crypt_get_dir(), action_argv[0],
+				ci == CRYPT_BUSY ? " and is in use" : "");
 		r = crypt_init_by_name(&cd, action_argv[0]);
 		if (r < 0 || !crypt_get_type(cd))
 			goto out;
 
-		log_std("  type:  %s\n", crypt_get_type(cd));
+		log_std("  type:    %s\n", crypt_get_type(cd));
 
 		r = crypt_get_active_device(cd, action_argv[0], &cad);
 		if (r < 0)
@@ -310,13 +420,21 @@ static int action_status(int arg)
 
 		log_std("  cipher:  %s-%s\n", crypt_get_cipher(cd), crypt_get_cipher_mode(cd));
 		log_std("  keysize: %d bits\n", crypt_get_volume_key_size(cd) * 8);
-		log_std("  device:  %s\n", crypt_get_device_name(cd));
+		device = crypt_get_device_name(cd);
+		log_std("  device:  %s\n", device);
+		if (crypt_loop_device(device)) {
+			backing_file = crypt_loop_backing_file(device);
+			log_std("  loop:    %s\n", backing_file);
+			free(backing_file);
+		}
 		log_std("  offset:  %" PRIu64 " sectors\n", cad.offset);
 		log_std("  size:    %" PRIu64 " sectors\n", cad.size);
 		if (cad.iv_offset)
 			log_std("  skipped: %" PRIu64 " sectors\n", cad.iv_offset);
 		log_std("  mode:    %s\n", cad.flags & CRYPT_ACTIVATE_READONLY ?
 					   "readonly" : "read/write");
+		if (cad.flags & CRYPT_ACTIVATE_ALLOW_DISCARDS)
+			log_std("  flags:   discards\n");
 	}
 out:
 	crypt_free(cd);
@@ -349,19 +467,24 @@ fail:
 	return -EINVAL;
 }
 
-static int action_luksFormat(int arg)
+static int action_luksFormat(int arg __attribute__((unused)))
 {
 	int r = -EINVAL, keysize;
+	const char *header_device;
 	char *msg = NULL, *key = NULL, cipher [MAX_CIPHER_LEN], cipher_mode[MAX_CIPHER_LEN];
 	char *password = NULL;
-	unsigned int passwordLen;
+	size_t passwordLen;
 	struct crypt_device *cd = NULL;
 	struct crypt_params_luks1 params = {
 		.hash = opt_hash ?: DEFAULT_LUKS1_HASH,
 		.data_alignment = opt_align_payload,
+		.data_device = opt_header_device ? action_argv[0] : NULL,
 	};
 
-	if(asprintf(&msg, _("This will overwrite data on %s irrevocably."), action_argv[0]) == -1) {
+	header_device = opt_header_device ?: action_argv[0];
+
+	if(asprintf(&msg, _("This will overwrite data on %s irrevocably."),
+		    header_device) == -1) {
 		log_err(_("memory allocation error in action_luksFormat"));
 		r = -ENOMEM;
 		goto out;
@@ -372,21 +495,24 @@ static int action_luksFormat(int arg)
 		goto out;
 
 	r = crypt_parse_name_and_mode(opt_cipher ?: DEFAULT_CIPHER(LUKS1),
-				      cipher, cipher_mode);
+				      cipher, NULL, cipher_mode);
 	if (r < 0) {
-		log_err("No known cipher specification pattern detected.\n");
+		log_err(_("No known cipher specification pattern detected.\n"));
 		goto out;
 	}
 
-	if ((r = crypt_init(&cd, action_argv[0])))
+	if ((r = crypt_init(&cd, header_device))) {
+		if (opt_header_device)
+			log_err(_("Cannot use %s as on-disk header.\n"), header_device);
 		goto out;
+	}
 
 	keysize = (opt_key_size ?: DEFAULT_LUKS1_KEYBITS) / 8;
 
 	crypt_set_password_verify(cd, 1);
 	crypt_set_timeout(cd, opt_timeout);
 	if (opt_iteration_time)
-		crypt_set_iterarion_time(cd, opt_iteration_time);
+		crypt_set_iteration_time(cd, opt_iteration_time);
 
 	if (opt_random)
 		crypt_set_rng_type(cd, CRYPT_RNG_RANDOM);
@@ -421,34 +547,57 @@ out:
 	return r;
 }
 
-static int action_luksOpen(int arg)
+static int action_luksOpen(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
+	const char *data_device, *header_device;
 	uint32_t flags = 0;
 	int r;
 
-	if ((r = crypt_init(&cd, action_argv[0])))
+	if (opt_header_device) {
+		header_device = opt_header_device;
+		data_device = action_argv[0];
+	} else {
+		header_device = action_argv[0];
+		data_device = NULL;
+	}
+
+	if ((r = crypt_init(&cd, header_device)))
 		goto out;
 
 	if ((r = crypt_load(cd, CRYPT_LUKS1, NULL)))
 		goto out;
 
+	if (data_device &&
+	    (r = crypt_set_data_device(cd, data_device)))
+		goto out;
+
+	if (!data_device && (crypt_get_data_offset(cd) < 8)) {
+		log_err(_("Reduced data offset is allowed only for detached LUKS header.\n"));
+		r = -EINVAL;
+		goto out;
+	}
+
 	crypt_set_timeout(cd, opt_timeout);
 	crypt_set_password_retry(cd, opt_tries);
 
 	if (opt_iteration_time)
-		crypt_set_iterarion_time(cd, opt_iteration_time);
+		crypt_set_iteration_time(cd, opt_iteration_time);
+
 	if (opt_readonly)
 		flags |= CRYPT_ACTIVATE_READONLY;
 
+	if (opt_allow_discards)
+		flags |= CRYPT_ACTIVATE_ALLOW_DISCARDS;
+
 	if (opt_key_file) {
 		crypt_set_password_retry(cd, 1);
 		r = crypt_activate_by_keyfile(cd, action_argv[1],
-			CRYPT_ANY_SLOT, opt_key_file, opt_keyfile_size,
+			opt_key_slot, opt_key_file, opt_keyfile_size,
 			flags);
 	} else
 		r = crypt_activate_by_passphrase(cd, action_argv[1],
-			CRYPT_ANY_SLOT, NULL, 0, flags);
+			opt_key_slot, NULL, 0, flags);
 out:
 	crypt_free(cd);
 	return r;
@@ -460,8 +609,8 @@ static int verify_keyslot(struct crypt_device *cd, int key_slot,
 {
 	crypt_keyslot_info ki;
 	char *password = NULL;
-	unsigned int passwordLen, i;
-	int r;
+	size_t passwordLen;
+	int i, r;
 
 	ki = crypt_keyslot_status(cd, key_slot);
 	if (ki == CRYPT_SLOT_ACTIVE_LAST && msg_last && !_yesDialog(msg_last, NULL))
@@ -498,7 +647,7 @@ out:
 	return r;
 }
 
-static int action_luksKillSlot(int arg)
+static int action_luksKillSlot(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r;
@@ -520,6 +669,7 @@ static int action_luksKillSlot(int arg)
 	case CRYPT_SLOT_INACTIVE:
 		log_err(_("Key %d not active. Can't wipe.\n"), opt_key_slot);
 	case CRYPT_SLOT_INVALID:
+		r = -EINVAL;
 		goto out;
 	}
 
@@ -538,11 +688,11 @@ out:
 	return r;
 }
 
-static int action_luksRemoveKey(int arg)
+static int action_luksRemoveKey(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	char *password = NULL;
-	unsigned int passwordLen;
+	size_t passwordLen;
 	int r;
 
 	if ((r = crypt_init(&cd, action_argv[0])))
@@ -586,7 +736,7 @@ out:
 	return r;
 }
 
-static int action_luksAddKey(int arg)
+static int action_luksAddKey(int arg __attribute__((unused)))
 {
 	int r = -EINVAL, keysize = 0;
 	char *key = NULL;
@@ -605,12 +755,13 @@ static int action_luksAddKey(int arg)
 	crypt_set_password_verify(cd, opt_verify_passphrase ? 1 : 0);
 	crypt_set_timeout(cd, opt_timeout);
 	if (opt_iteration_time)
-		crypt_set_iterarion_time(cd, opt_iteration_time);
+		crypt_set_iteration_time(cd, opt_iteration_time);
 
 	if (opt_master_key_file) {
-		if (_read_mk(opt_master_key_file, &key, keysize) < 0)
+		r = _read_mk(opt_master_key_file, &key, keysize);
+		if (r < 0)
 			goto out;
-
+		//FIXME: process keyfile arg
 		r = crypt_keyslot_add_by_volume_key(cd, opt_key_slot,
 						    key, keysize, NULL, 0);
 	} else if (opt_key_file || opt_new_key_file) {
@@ -627,7 +778,104 @@ out:
 	return r;
 }
 
-static int action_isLuks(int arg)
+static int _slots_full(struct crypt_device *cd)
+{
+	int i;
+
+	for (i = 0; i < crypt_keyslot_max(crypt_get_type(cd)); i++)
+		if (crypt_keyslot_status(cd, i) == CRYPT_SLOT_INACTIVE)
+			return 0;
+	return 1;
+}
+
+static int action_luksChangeKey(int arg __attribute__((unused)))
+{
+	const char *opt_new_key_file = (action_argc > 1 ? action_argv[1] : NULL);
+	struct crypt_device *cd = NULL;
+	char *vk = NULL, *password = NULL;
+	size_t passwordLen = 0;
+	size_t vk_size;
+	int new_key_slot, old_key_slot, r;
+
+	if ((r = crypt_init(&cd, action_argv[0])))
+		goto out;
+
+	if ((r = crypt_load(cd, CRYPT_LUKS1, NULL)))
+		goto out;
+
+	if (opt_iteration_time)
+		crypt_set_iteration_time(cd, opt_iteration_time);
+
+	r = crypt_get_key(_("Enter LUKS passphrase to be changed: "),
+		      &password, &passwordLen,
+		      opt_keyfile_size, opt_key_file, opt_timeout,
+		      opt_batch_mode ? 0 : opt_verify_passphrase, cd);
+	if (r < 0)
+		goto out;
+
+	vk_size = crypt_get_volume_key_size(cd);
+	vk = crypt_safe_alloc(vk_size);
+	if (!vk) {
+		r = -ENOMEM;
+		goto out;
+	}
+
+	r = crypt_volume_key_get(cd, opt_key_slot, vk, &vk_size,
+				 password, passwordLen);
+	if (r < 0) {
+		if (opt_key_slot != CRYPT_ANY_SLOT)
+			log_err(_("No key available with this passphrase.\n"));
+		goto out;
+	}
+
+	if (opt_key_slot != CRYPT_ANY_SLOT || _slots_full(cd)) {
+		log_dbg("Key slot %d is going to be overwritten (%s).",
+			r, opt_key_slot != CRYPT_ANY_SLOT ?
+			"explicit key slot specified" : "no free key slot");
+		old_key_slot = r;
+		new_key_slot = r;
+	} else {
+		log_dbg("Allocating new key slot.");
+		old_key_slot = r;
+		new_key_slot = CRYPT_ANY_SLOT;
+	}
+
+	crypt_safe_free(password);
+	password = NULL;
+	passwordLen = 0;
+	r = crypt_get_key(_("Enter new LUKS passphrase: "),
+			  &password, &passwordLen,
+			  opt_new_keyfile_size, opt_new_key_file,
+			  opt_timeout, opt_batch_mode ? 0 : 1, cd);
+	if (r < 0)
+		goto out;
+
+	if (new_key_slot == old_key_slot) {
+		(void)crypt_keyslot_destroy(cd, old_key_slot);
+		r = crypt_keyslot_add_by_volume_key(cd, new_key_slot,
+						    vk, vk_size,
+						    password, passwordLen);
+		if (r >= 0)
+			log_verbose(_("Key slot %d changed.\n"), r);
+	} else {
+		r = crypt_keyslot_add_by_volume_key(cd, CRYPT_ANY_SLOT,
+						    vk, vk_size,
+						    password, passwordLen);
+		if (r >= 0) {
+			log_verbose(_("Replaced with key slot %d.\n"), r);
+			r = crypt_keyslot_destroy(cd, old_key_slot);
+		}
+	}
+	if (r < 0)
+		log_err(_("Failed to swap new key slot.\n"));
+out:
+	crypt_safe_free(vk);
+	crypt_safe_free(password);
+	crypt_free(cd);
+	return r;
+}
+
+static int action_isLuks(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r;
@@ -635,13 +883,14 @@ static int action_isLuks(int arg)
 	if ((r = crypt_init(&cd, action_argv[0])))
 		goto out;
 
+	crypt_set_log_callback(cd, _quiet_log, NULL);
 	r = crypt_load(cd, CRYPT_LUKS1, NULL);
 out:
 	crypt_free(cd);
 	return r;
 }
 
-static int action_luksUUID(int arg)
+static int action_luksUUID(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	const char *existing_uuid = NULL;
@@ -670,9 +919,10 @@ out:
 static int luksDump_with_volume_key(struct crypt_device *cd)
 {
 	char *vk = NULL, *password = NULL;
-	unsigned int passwordLen = 0;
+	size_t passwordLen = 0;
 	size_t vk_size;
-	int i, r;
+	unsigned i;
+	int r;
 
 	crypt_set_confirm_callback(cd, _yesDialog, NULL);
 	if (!_yesDialog(
@@ -700,9 +950,9 @@ static int luksDump_with_volume_key(struct crypt_device *cd)
 	log_std("LUKS header information for %s\n", crypt_get_device_name(cd));
 	log_std("Cipher name:   \t%s\n", crypt_get_cipher(cd));
 	log_std("Cipher mode:   \t%s\n", crypt_get_cipher_mode(cd));
-	log_std("Payload offset:\t%d\n", crypt_get_data_offset(cd));
+	log_std("Payload offset:\t%d\n", (int)crypt_get_data_offset(cd));
 	log_std("UUID:          \t%s\n", crypt_get_uuid(cd));
-	log_std("MK bits:       \t%d\n", vk_size * 8);
+	log_std("MK bits:       \t%d\n", (int)vk_size * 8);
 	log_std("MK dump:\t");
 
 	for(i = 0; i < vk_size; i++) {
@@ -718,7 +968,7 @@ out:
 	return r;
 }
 
-static int action_luksDump(int arg)
+static int action_luksDump(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r;
@@ -738,12 +988,12 @@ out:
 	return r;
 }
 
-static int action_luksSuspend(int arg)
+static int action_luksSuspend(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r;
 
-	r = crypt_init_by_name(&cd, action_argv[0]);
+	r = crypt_init_by_name_and_header(&cd, action_argv[0], opt_header_device);
 	if (!r)
 		r = crypt_suspend(cd, action_argv[0]);
 
@@ -751,16 +1001,16 @@ static int action_luksSuspend(int arg)
 	return r;
 }
 
-static int action_luksResume(int arg)
+static int action_luksResume(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r;
 
-	if ((r = crypt_init_by_name(&cd, action_argv[0])))
+	if ((r = crypt_init_by_name_and_header(&cd, action_argv[0], opt_header_device)))
 		goto out;
 
-	if ((r = crypt_load(cd, CRYPT_LUKS1, NULL)))
-		goto out;
+	crypt_set_timeout(cd, opt_timeout);
+	crypt_set_password_retry(cd, opt_tries);
 
 	if (opt_key_file)
 		r = crypt_resume_by_keyfile(cd, action_argv[0], CRYPT_ANY_SLOT,
@@ -773,7 +1023,7 @@ out:
 	return r;
 }
 
-static int action_luksBackup(int arg)
+static int action_luksBackup(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r;
@@ -794,7 +1044,7 @@ out:
 	return r;
 }
 
-static int action_luksRestore(int arg)
+static int action_luksRestore(int arg __attribute__((unused)))
 {
 	struct crypt_device *cd = NULL;
 	int r = 0;
@@ -821,11 +1071,15 @@ static __attribute__ ((noreturn)) void usage(poptContext popt_context,
 	poptPrintUsage(popt_context, stderr, 0);
 	if (error)
 		log_err("%s: %s\n", more, error);
+	poptFreeContext(popt_context);
 	exit(exitcode);
 }
 
-static void help(poptContext popt_context, enum poptCallbackReason reason,
-                 struct poptOption *key, const char * arg, void *data)
+static void help(poptContext popt_context,
+		 enum poptCallbackReason reason __attribute__((unused)),
+		 struct poptOption *key,
+		 const char *arg __attribute__((unused)),
+		 void *data __attribute__((unused)))
 {
 	if (key->shortName == '?') {
 		struct action_type *action;
@@ -847,9 +1101,16 @@ static void help(poptContext popt_context, enum poptCallbackReason reason,
 			 "<key file> optional key file for the new key for luksAddKey action\n"),
 			crypt_get_dir());
 
+		log_std(_("\nDefault compiled-in keyfile parameters:\n"
+			 "\tMaximum keyfile size: %dkB, "
+			 "Maximum interactive passphrase length %d (characters)\n"),
+			 DEFAULT_KEYFILE_SIZE_MAXKB, DEFAULT_PASSPHRASE_SIZE_MAX);
+
 		log_std(_("\nDefault compiled-in device cipher parameters:\n"
+			 "\tloop-AES: %s, Key %d bits\n"
 			 "\tplain: %s, Key: %d bits, Password hashing: %s\n"
 			 "\tLUKS1: %s, Key: %d bits, LUKS header hashing: %s, RNG: %s\n"),
+			 DEFAULT_LOOPAES_CIPHER, DEFAULT_LOOPAES_KEYBITS,
 			 DEFAULT_CIPHER(PLAIN), DEFAULT_PLAIN_KEYBITS, DEFAULT_PLAIN_HASH,
 			 DEFAULT_CIPHER(LUKS1), DEFAULT_LUKS1_KEYBITS, DEFAULT_LUKS1_HASH,
 			 DEFAULT_RNG);
@@ -858,7 +1119,7 @@ static void help(poptContext popt_context, enum poptCallbackReason reason,
 		usage(popt_context, EXIT_SUCCESS, NULL, NULL);
 }
 
-static void _dbg_version_and_cmd(int argc, char **argv)
+static void _dbg_version_and_cmd(int argc, const char **argv)
 {
 	int i;
 
@@ -866,7 +1127,7 @@ static void _dbg_version_and_cmd(int argc, char **argv)
 	for (i = 0; i < argc; i++) {
 		if (i)
 			log_std(" ");
-		log_std(argv[i]);
+		log_std("%s", argv[i]);
 	}
 	log_std("\"\n");
 }
@@ -875,6 +1136,8 @@ static int run_action(struct action_type *action)
 {
 	int r;
 
+	log_dbg("Running command %s.", action->type);
+
 	if (action->required_memlock)
 		crypt_memory_lock(NULL, 1);
 
@@ -906,7 +1169,7 @@ static int run_action(struct action_type *action)
 	return r;
 }
 
-int main(int argc, char **argv)
+int main(int argc, const char **argv)
 {
 	static char *popt_tmp;
 	static struct poptOption popt_help_options[] = {
@@ -927,8 +1190,8 @@ int main(int argc, char **argv)
 		{ "master-key-file",  '\0',  POPT_ARG_STRING, &opt_master_key_file,     0, N_("Read the volume (master) key from file."), NULL },
 		{ "dump-master-key",  '\0',  POPT_ARG_NONE, &opt_dump_master_key,       0, N_("Dump volume (master) key instead of keyslots info."), NULL },
 		{ "key-size",          's',  POPT_ARG_INT, &opt_key_size,               0, N_("The size of the encryption key"), N_("BITS") },
-		{ "keyfile-size",      'l',  POPT_ARG_INT, &opt_keyfile_size,           0, N_("Limits the read from keyfile"), N_("bytes") },
-		{ "new-keyfile-size", '\0',  POPT_ARG_INT, &opt_new_keyfile_size,       0, N_("Limits the read from newly added keyfile"), N_("bytes") },
+		{ "keyfile-size",      'l',  POPT_ARG_LONG, &opt_keyfile_size,          0, N_("Limits the read from keyfile"), N_("bytes") },
+		{ "new-keyfile-size", '\0',  POPT_ARG_LONG, &opt_new_keyfile_size,      0, N_("Limits the read from newly added keyfile"), N_("bytes") },
 		{ "key-slot",          'S',  POPT_ARG_INT, &opt_key_slot,               0, N_("Slot number for new key (default is first free)"), NULL },
 		{ "size",              'b',  POPT_ARG_STRING, &popt_tmp,                1, N_("The size of the device"), N_("SECTORS") },
 		{ "offset",            'o',  POPT_ARG_STRING, &popt_tmp,                2, N_("The start offset in the backend device"), N_("SECTORS") },
@@ -942,12 +1205,15 @@ int main(int argc, char **argv)
 		{ "header-backup-file",'\0', POPT_ARG_STRING, &opt_header_backup_file,  0, N_("File with LUKS header and keyslots backup."), NULL },
 		{ "use-random",        '\0', POPT_ARG_NONE, &opt_random,                0, N_("Use /dev/random for generating volume key."), NULL },
 		{ "use-urandom",       '\0', POPT_ARG_NONE, &opt_urandom,               0, N_("Use /dev/urandom for generating volume key."), NULL },
-		{ "uuid",              '\0',  POPT_ARG_STRING, &opt_uuid,               0, N_("UUID for device to use."), NULL },
+		{ "shared",            '\0', POPT_ARG_NONE, &opt_shared,                0, N_("Share device with another non-overlapping crypt segment."), NULL },
+		{ "uuid",              '\0', POPT_ARG_STRING, &opt_uuid,                0, N_("UUID for device to use."), NULL },
+		{ "allow-discards",    '\0', POPT_ARG_NONE, &opt_allow_discards,        0, N_("Allow discards (aka TRIM) requests for device."), NULL },
+		{ "header",            '\0', POPT_ARG_STRING, &opt_header_device,       0, N_("Device or file with separated LUKS header."), NULL },
 		POPT_TABLEEND
 	};
 	poptContext popt_context;
 	struct action_type *action;
-	char *aname;
+	const char *aname;
 	int r;
 	const char *null_action_argv[] = {NULL};
 
@@ -957,8 +1223,7 @@ int main(int argc, char **argv)
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 
-	popt_context = poptGetContext(PACKAGE, argc, (const char **)argv,
-	                              popt_options, 0);
+	popt_context = poptGetContext(PACKAGE, argc, argv, popt_options, 0);
 	poptSetOtherOptionHelp(popt_context,
 	                       N_("[OPTION...] <action> <action-specific>]"));
 
@@ -966,8 +1231,11 @@ int main(int argc, char **argv)
 		unsigned long long ull_value;
 		char *endp;
 
+		errno = 0;
 		ull_value = strtoull(popt_tmp, &endp, 0);
-		if (*endp || !*popt_tmp)
+		if (*endp || !*popt_tmp ||
+		    (errno == ERANGE && ull_value == ULLONG_MAX) ||
+		    (errno != 0 && ull_value == 0))
 			r = POPT_ERROR_BADNUMBER;
 
 		switch(r) {
@@ -979,6 +1247,7 @@ int main(int argc, char **argv)
 				break;
 			case 3:
 				opt_skip = ull_value;
+				opt_skip_valid = 1;
 				break;
 		}
 
@@ -991,10 +1260,11 @@ int main(int argc, char **argv)
 		      poptBadOption(popt_context, POPT_BADOPTION_NOALIAS));
 	if (opt_version_mode) {
 		log_std("%s %s\n", PACKAGE_NAME, PACKAGE_VERSION);
+		poptFreeContext(popt_context);
 		exit(EXIT_SUCCESS);
 	}
 
-	if (!(aname = (char *)poptGetArg(popt_context)))
+	if (!(aname = poptGetArg(popt_context)))
 		usage(popt_context, EXIT_FAILURE, _("Argument <action> missing."),
 		      poptGetInvocationName(popt_context));
 	for(action = action_types; action->type; action++)
@@ -1023,11 +1293,27 @@ int main(int argc, char **argv)
 
 	/* FIXME: rewrite this from scratch */
 
+	if (opt_shared && strcmp(aname, "create")) {
+		usage(popt_context, EXIT_FAILURE,
+		      _("Option --shared is allowed only for create operation.\n"),
+		      poptGetInvocationName(popt_context));
+	}
+
+	if (opt_allow_discards &&
+	    strcmp(aname, "luksOpen") &&
+	    strcmp(aname, "create") &&
+	    strcmp(aname, "loopaesOpen")) {
+		usage(popt_context, EXIT_FAILURE,
+		      _("Option --allow-discards is allowed only for luksOpen, loopaesOpen and create operation.\n"),
+		      poptGetInvocationName(popt_context));
+	}
+
 	if (opt_key_size &&
 	   strcmp(aname, "luksFormat") &&
-	   strcmp(aname, "create")) {
+	   strcmp(aname, "create") &&
+	   strcmp(aname, "loopaesOpen")) {
 		usage(popt_context, EXIT_FAILURE,
-		      _("Option --key-size is allowed only for luksFormat and create.\n"
+		      _("Option --key-size is allowed only for luksFormat, create and loopaesOpen.\n"
 		        "To limit read from keyfile use --keyfile-size=(bytes)."),
 		      poptGetInvocationName(popt_context));
 	}
@@ -1040,7 +1326,7 @@ int main(int argc, char **argv)
 	if (!strcmp(aname, "luksKillSlot") && action_argc > 1)
 		opt_key_slot = atoi(action_argv[1]);
 	if (opt_key_slot != CRYPT_ANY_SLOT &&
-	    (opt_key_slot < 0 || opt_key_slot > crypt_keyslot_max(CRYPT_LUKS1)))
+	    (opt_key_slot < 0 || opt_key_slot >= crypt_keyslot_max(CRYPT_LUKS1)))
 		usage(popt_context, EXIT_FAILURE, _("Key slot is invalid."),
 		      poptGetInvocationName(popt_context));
 
@@ -1050,7 +1336,13 @@ int main(int argc, char **argv)
 		if (opt_key_file)
 			log_err(_("Option --key-file takes precedence over specified key file argument.\n"));
 		else
-			opt_key_file = (char*)action_argv[1];
+			opt_key_file = action_argv[1];
+	}
+
+	if (opt_keyfile_size < 0 || opt_new_keyfile_size < 0 || opt_key_size < 0) {
+		usage(popt_context, EXIT_FAILURE,
+		      _("Negative number for option not permitted."),
+		      poptGetInvocationName(popt_context));
 	}
 
 	if (opt_random && opt_urandom)
@@ -1064,9 +1356,15 @@ int main(int argc, char **argv)
 		usage(popt_context, EXIT_FAILURE, _("Option --uuid is allowed only for luksFormat and luksUUID."),
 		      poptGetInvocationName(popt_context));
 
-	if ((opt_offset || opt_skip) && strcmp(aname, "create"))
-		usage(popt_context, EXIT_FAILURE, _("Options --offset and --skip are supported only for create command.\n"),
-		      poptGetInvocationName(popt_context));
+	if (opt_skip && strcmp(aname, "create") && strcmp(aname, "loopaesOpen"))
+		usage(popt_context, EXIT_FAILURE,
+		_("Option --skip is supported only for create and loopaesOpen commands.\n"),
+		poptGetInvocationName(popt_context));
+
+	if (opt_offset && strcmp(aname, "create") && strcmp(aname, "loopaesOpen"))
+		usage(popt_context, EXIT_FAILURE,
+		_("Option --offset is supported only for create and loopaesOpen commands.\n"),
+		poptGetInvocationName(popt_context));
 
 	if (opt_debug) {
 		opt_verbose = 1;
@@ -1074,5 +1372,7 @@ int main(int argc, char **argv)
 		_dbg_version_and_cmd(argc, argv);
 	}
 
-	return run_action(action);
+	r = run_action(action);
+	poptFreeContext(popt_context);
+	return r;
 }
